/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.io.internal;

import java.util.AbstractQueue;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessage;

/**
 * <p>
 * Queue for holding {@link ChannelAccessMessage}s.
 * </p>
 * 
 * <p>
 * The messages are automatically ordered by their priority: CA_PROTO_VERSION,
 * CA_PROTO_CLIENT_NAME, and CA_PROTO_HOST_NAME messages have the highest
 * priority. CA_PROTO_EVENTS_OFF and CA_PROTO_EVENTS_ON messages have the second
 * highest priority. CA_PROTO_ECHO messages have the third highest priority. All
 * other messages have the lowest priority. Messages with a higher priority are
 * close to the head of the queue (returned earlier), messages with lower
 * priority are closer to the tail of the queue (returned later). Messages of
 * the same priority have the order they were added in (messages that were added
 * earlier are closer to the head of the queue).
 * </p>
 * 
 * <p>
 * This queue has been specifically designed for holding messages that are going
 * to be sent. Some messages should be sent with a higher priority than others
 * (e.g. CA_PROTO_VERSION messages must be at the start of a connection,
 * CA_PROTO_ECHO messages are more important than most other messages because
 * timing is important for them).
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ChannelAccessMessageQueue extends
        AbstractQueue<ChannelAccessMessage> {

    private LinkedList<ChannelAccessMessage> versionUserNameHostNameMessages = new LinkedList<ChannelAccessMessage>();
    private LinkedList<ChannelAccessMessage> eventsOnOffMessages = new LinkedList<ChannelAccessMessage>();
    private LinkedList<ChannelAccessMessage> echoMessages = new LinkedList<ChannelAccessMessage>();
    private LinkedList<ChannelAccessMessage> otherMessages = new LinkedList<ChannelAccessMessage>();

    private LinkedList<LinkedList<ChannelAccessMessage>> allMessageQueues = new LinkedList<LinkedList<ChannelAccessMessage>>();

    private int size;

    /**
     * Creates an empty queue.
     */
    public ChannelAccessMessageQueue() {
        allMessageQueues.add(versionUserNameHostNameMessages);
        allMessageQueues.add(eventsOnOffMessages);
        allMessageQueues.add(echoMessages);
        allMessageQueues.add(otherMessages);
    }

    @Override
    public boolean offer(ChannelAccessMessage message) {
        // We could support more than 2^31-1 messages in the queue, but there
        // is no good reason to queue so many messages and this way the
        // implementation is simple and fast.
        if (size == Integer.MAX_VALUE) {
            return false;
        }
        switch (message.getCommand()) {
        case CA_PROTO_VERSION:
        case CA_PROTO_CLIENT_NAME:
        case CA_PROTO_HOST_NAME:
            versionUserNameHostNameMessages.add(message);
            break;
        case CA_PROTO_EVENTS_OFF:
        case CA_PROTO_EVENTS_ON:
            eventsOnOffMessages.add(message);
            break;
        case CA_PROTO_ECHO:
            echoMessages.add(message);
            break;
        default:
            otherMessages.add(message);
        }
        size++;
        return true;
    }

    @Override
    public ChannelAccessMessage poll() {
        for (LinkedList<ChannelAccessMessage> queue : allMessageQueues) {
            ChannelAccessMessage message = queue.poll();
            if (message != null) {
                size--;
                return message;
            }
        }
        return null;
    }

    @Override
    public ChannelAccessMessage peek() {
        for (LinkedList<ChannelAccessMessage> queue : allMessageQueues) {
            ChannelAccessMessage message = queue.peek();
            if (message != null) {
                return message;
            }
        }
        return null;
    }

    @Override
    public Iterator<ChannelAccessMessage> iterator() {
        return new Iterator<ChannelAccessMessage>() {

            private Iterator<LinkedList<ChannelAccessMessage>> queueIterator = allMessageQueues
                    .iterator();
            private Iterator<ChannelAccessMessage> messageIterator = null;
            private Iterator<ChannelAccessMessage> oldMessageIterator = null;

            @Override
            public boolean hasNext() {
                while (messageIterator == null || !messageIterator.hasNext()) {
                    if (!queueIterator.hasNext()) {
                        return false;
                    } else {
                        messageIterator = queueIterator.next().iterator();
                    }
                }
                return true;
            }

            @Override
            public ChannelAccessMessage next() {
                if (!hasNext()) {
                    throw new NoSuchElementException();
                }
                // We remember the iterator that has been used for retrieving
                // the message so that we can later use its remove method. The
                // messageIterator reference might be replaced by a different
                // iterator when hasNext is called.
                oldMessageIterator = messageIterator;
                return messageIterator.next();
            }

            @Override
            public void remove() {
                if (oldMessageIterator == null) {
                    throw new IllegalStateException(
                            "next() must be called before calling remove().");
                }
                oldMessageIterator.remove();
            }

        };
    }

    @Override
    public int size() {
        return size;
    }

    @Override
    public void clear() {
        // We do not have to implement this explicitly, but our implementation
        // should be much faster than the default one.
        for (LinkedList<ChannelAccessMessage> queue : allMessageQueues) {
            queue.clear();
        }
        size = 0;
    }

    /**
     * Returns a string representation of this queue. The string representation
     * is built by concatenating the string representation of all the messages
     * that are stored in the queue. The messages are concatenated in the same
     * order in which they would be returned by this queue.
     * 
     * @return string representation of this queue.
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(this.getClass().getName());
        sb.append('[');
        boolean first = true;
        for (ChannelAccessMessage message : this) {
            if (!first) {
                sb.append(", ");
            } else {
                first = false;
            }
            sb.append(message);
        }
        sb.append(']');
        return sb.toString();
    }

    /**
     * <p>
     * Returns a hash code value for this queue. This method is supported for
     * the benefit of hash tables. The hash code is calculated by combining the
     * hash codes of all messages that are stored in the queue.
     * </p>
     * 
     * <p>
     * If two queues are equal (as indicated by the return value of
     * {@link #equals(Object)}), they also have the same hash code. However, the
     * fact that two queues have the same hash code does not imply that they are
     * equal.
     * </p>
     * 
     * @return hash code suitable for use by a hash table.
     */
    @Override
    public int hashCode() {
        return new HashCodeBuilder().append(versionUserNameHostNameMessages)
                .append(eventsOnOffMessages).append(echoMessages)
                .append(otherMessages).toHashCode();
    }

    /**
     * Tells whether the specified object represents an equal queue. Two queues
     * are considered equal if they contain the same messages in the same order.
     * 
     * @param obj
     *            reference object with which this object is compared.
     * @return <code>true</code> if <code>obj</code> is a queue that contains
     *         the same messages in the same order as this queue,
     *         <code>false</code> otherwise.
     */
    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof ChannelAccessMessageQueue)) {
            return false;
        }
        ChannelAccessMessageQueue other = (ChannelAccessMessageQueue) obj;
        return new EqualsBuilder()
                .append(this.versionUserNameHostNameMessages,
                        other.versionUserNameHostNameMessages)
                .append(this.eventsOnOffMessages, other.eventsOnOffMessages)
                .append(this.echoMessages, other.echoMessages)
                .append(this.otherMessages, other.otherMessages).isEquals();
    }

}
