/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver.internal;

import com.aquenos.epics.jackie.client.resolver.ChannelNameResolver.ChannelNameSearchResultProcessor;

/**
 * Channel-name search request. This class bundles all information related to a
 * specific search request. Most of the fields are immutable and are initialized
 * from the parameters given to the channel name resolver. The search timer,
 * however, is initialized with <code>null</code> and can be updated by the
 * {@link DefaultChannelNameResolverImpl} when it adds the request to a certain
 * search timer or moves it to a different timer.
 * 
 * @author Sebastian Marsching
 */
public class ChannelSearchRequest {

    private String channelName;
    private int cid;
    private ChannelNameSearchResultProcessor resultProcessor;
    private SearchTimer searchTimer;

    /**
     * Creates a channel-name search request. The search timer is initialized
     * with <code>null</code>.
     * 
     * @param channelName
     *            channel name.
     * @param cid
     *            channel CID (assigned by the code calling the channel name
     *            resolver).
     * @param resultProcessor
     *            result processor that is notified when the channel name could
     *            be resolved.
     */
    ChannelSearchRequest(String channelName, int cid,
            ChannelNameSearchResultProcessor resultProcessor) {
        this.channelName = channelName;
        this.cid = cid;
        this.resultProcessor = resultProcessor;
    }

    /**
     * Returns the channel name.
     * 
     * @return channel name.
     */
    public String getChannelName() {
        return channelName;
    }

    /**
     * Returns the client-side numeric ID for the channel.
     * 
     * @return client-assigned channel ID.
     */
    public int getCID() {
        return cid;
    }

    /**
     * Returns the result processor that is notified when the channel name could
     * be resolved.
     * 
     * @return result processor.
     */
    public ChannelNameSearchResultProcessor getResultProcessor() {
        return resultProcessor;
    }

    /**
     * Returns the current search timer for the request. Might return
     * <code>null</code> if the request has not been added to a search timer
     * yet.
     * 
     * @return current search timer or <code>null</code>.
     */
    public SearchTimer getSearchTimer() {
        return searchTimer;
    }

    /**
     * Sets the current search timer for the request. This method should only be
     * called by {@link DefaultChannelNameResolverImpl} when it assigns the
     * request to a search timer.
     * 
     * @param searchTimer
     *            search timer that is now handling this request.
     */
    public void setSearchTimer(SearchTimer searchTimer) {
        this.searchTimer = searchTimer;
    }

}
