/*
 * Copyright 2016-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.util.LinkedList;
import java.util.Random;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.junit.Test;

import com.aquenos.epics.jackie.common.exception.NullErrorHandler;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.util.ListenableFuture;
import com.aquenos.epics.jackie.common.util.ListenerLockPolicy;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessPuttableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Tests for the {@link DefaultClientThreadingStrategy}.
 * 
 * @author Sebastian Marsching
 */
public class DefaultClientThreadingStrategyTest {

    /**
     * Simple stub implementation of a {@link ChannelAccessChannel}. This
     * implementation only implements the <code>getName()</code> method because
     * this is the only one that might be called by the
     * {@link DefaultClientThreadingStrategy}.
     * 
     * @author Sebastian Marsching
     */
    private static class ChannelAccessChannelStub
            extends AbstractChannelAccessChannel {

        public ChannelAccessChannelStub() {
            super(ListenerLockPolicy.BLOCK);
        }

        @Override
        public void destroy() {
            throw new UnsupportedOperationException();
        }

        @Override
        public String getName() {
            return "testChannel";
        }

        @Override
        public ChannelAccessClient getClient() {
            throw new UnsupportedOperationException();
        }

        @Override
        public int getNativeCount() {
            throw new UnsupportedOperationException();
        }

        @Override
        public ChannelAccessValueType getNativeDataType() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean isDestroyed() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean isMayRead() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean isMayWrite() {
            throw new UnsupportedOperationException();
        }

        @Override
        public ListenableFuture<? extends ChannelAccessGettableValue<?>> get(
                ChannelAccessValueType valueType, int count) {
            throw new UnsupportedOperationException();
        }

        @Override
        public ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
                ChannelAccessValueType valueType, int count,
                ChannelAccessEventMask eventMask) {
            throw new UnsupportedOperationException();
        }

        @Override
        public ListenableFuture<Void> put(ChannelAccessPuttableValue<?> value) {
            throw new UnsupportedOperationException();
        }

    }

    /**
     * Channel connection listener internally used for tests. This connection
     * listener simply keeps record of the notifications it received. In
     * addition to that, it can sleep for a configurable amount of time each
     * time it is invoked.
     * 
     * @author Sebastian Marsching
     */
    private static class ChannelAccessConnectionListenerTestImpl
            implements ChannelAccessConnectionListener {

        public long sleepTime;
        public LinkedList<Boolean> recordedConnectionStateValues = new LinkedList<Boolean>();

        @Override
        public void connectionStateChanged(ChannelAccessChannel channel,
                boolean nowConnected) {
            synchronized (recordedConnectionStateValues) {
                recordedConnectionStateValues
                        .add(Boolean.valueOf(nowConnected));
            }
            if (sleepTime > 0L) {
                try {
                    Thread.sleep(sleepTime);
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
            }
        }

    }

    /**
     * Tests that connection listener notifications work as expected. In
     * particular, this test verifies that a listener does not receive out of
     * order notifications when there is more than one thread processing
     * notifications. While this method only tests notifications for a
     * connection listener, this result should also be valid for monitor
     * listeners because internally, they use the same implementation.
     */
    @Test
    public void testConnectionListenerNotifications() {
        ExecutorService executorService = Executors.newFixedThreadPool(10);
        DefaultClientThreadingStrategy threadingStrategy = new DefaultClientThreadingStrategy(
                new NullErrorHandler(), executorService, false);
        // First, we test with neither the sender nor the receiver sleeping.
        testConnectionListenerNotifications(threadingStrategy, 0L, 0L);
        // Next, we test with only the receiver sleeping.
        testConnectionListenerNotifications(threadingStrategy, 2L, 0L);
        // Next, we test with only the sender sleeping.
        testConnectionListenerNotifications(threadingStrategy, 0L, 2L);
        // Finally, we test with both the sender and receiver sleeping.
        testConnectionListenerNotifications(threadingStrategy, 2L, 2L);
        // Now that we are finished, we can destroy the threading strategy. The
        // executor service shut not be shutdown because we specified false for
        // the corresponding parameter.
        threadingStrategy.destroy();
        assertFalse(executorService.isShutdown());
        // Now we try the same again, but this time with the parameter being
        // true.
        threadingStrategy = new DefaultClientThreadingStrategy(
                new NullErrorHandler(), executorService, true);
        threadingStrategy.destroy();
        assertTrue(executorService.isShutdown());
        // Finally, we want to make sure that the threading strategy also works
        // correctly when no executor service is specified (the default case).
        // we do not have to make tests with sleeping because concurrency should
        // not be an issue in this case.
        threadingStrategy = new DefaultClientThreadingStrategy(
                new NullErrorHandler());
        testConnectionListenerNotifications(threadingStrategy, 0L, 0L);
    }

    private void testConnectionListenerNotifications(
            DefaultClientThreadingStrategy threadingStrategy,
            long receiverSleepTime, long senderSleepTime) {
        receiverSleepTime = Math.max(receiverSleepTime, 0L);
        senderSleepTime = Math.max(senderSleepTime, 0L);
        ChannelAccessChannel channel = new ChannelAccessChannelStub();
        ChannelAccessConnectionListenerTestImpl listener = new ChannelAccessConnectionListenerTestImpl();
        listener.sleepTime = receiverSleepTime;
        Random random = new Random();
        LinkedList<Boolean> expectedConnectionStateValues = new LinkedList<Boolean>();
        for (int i = 0; i < 1000; ++i) {
            expectedConnectionStateValues
                    .add(Boolean.valueOf(random.nextBoolean()));
        }
        for (Boolean connectionState : expectedConnectionStateValues) {
            threadingStrategy.notifyConnectionListener(channel, listener,
                    connectionState);
            if (senderSleepTime > 0L) {
                try {
                    Thread.sleep(senderSleepTime);
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
            }
        }
        // As notifications are asynchronous, it can take some time until the
        // last notification arrives. If the receiving thread sleeps some time
        // for each notification, this might take even longer.
        long deadline = System.currentTimeMillis() + 5000L
                + (receiverSleepTime > senderSleepTime
                        ? ((receiverSleepTime - senderSleepTime)
                                * expectedConnectionStateValues.size())
                        : 0);
        while (!Thread.currentThread().isInterrupted()) {
            synchronized (listener.recordedConnectionStateValues) {
                if (listener.recordedConnectionStateValues
                        .size() == expectedConnectionStateValues.size()
                        || System.currentTimeMillis() >= deadline) {
                    break;
                }
            }
            try {
                Thread.sleep(20);
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            }
        }
        synchronized (listener.recordedConnectionStateValues) {
            assertEquals(expectedConnectionStateValues,
                    listener.recordedConnectionStateValues);
        }
    }

}
