/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import java.io.IOException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.spi.SelectorProvider;
import java.util.Comparator;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 * <p>
 * Simple communication controller implementation.
 * </p>
 * 
 * <p>
 * Code using his implementation should call the {@link #processTimersAndIO()}
 * from a loop. This method processes all timers that are due and then uses a
 * {@link Selector} to wait for one of five things to happen:
 * </p>
 * <ul>
 * <li>A timeout is reached. The timeout is chosen so that it should occur when
 * the next timer event is due.</li>
 * <li>One of the channels that was registered with this controller is ready.</li>
 * <li>This object's {@link #wakeUp()} method is called.</li>
 * <li>This object's
 * {@link #registerChannel(ChannelProcessor, SelectableChannel)} method is
 * called.</li>
 * <li>This object's {@link #registerTimer(TimerProcessor, long)} method is
 * called.</li>
 * </ul>
 * 
 * <p>
 * After the selection operation has finished, all processors that have
 * registered channels that are ready now are processed. After that, the method
 * returns.
 * </p>
 * 
 * <p>
 * If one of the {@link ChannelProcessor#processIO() processIO()} or
 * {@link TimerProcessor#processTimer() processTimer()} methods throws an
 * exception, the processing is stopped immediately and the exception is passed
 * up to the code calling {@link #processTimersAndIO()}. This code should catch
 * the exception, handle it appropriately and then call
 * {@link #processTimersAndIO()} again.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class SimpleCommunicationController implements CommunicationController {

    private static class TimerEvent {

        private long eventTime;
        private TimerProcessor processor;

        private TimerEvent(long eventTime, TimerProcessor processor) {
            this.eventTime = eventTime;
            this.processor = processor;
        }

        public long getEventTime() {
            return eventTime;
        }

        public TimerProcessor getProcessor() {
            return processor;
        }

    }

    private static class WrappingSelectionKey extends SelectionKey {

        private SelectionKey wrappedSelectionKey;

        private WrappingSelectionKey(SelectionKey wrappedSelectionKey) {
            this.wrappedSelectionKey = wrappedSelectionKey;
        }

        @Override
        public SelectableChannel channel() {
            return wrappedSelectionKey.channel();
        }

        @Override
        public Selector selector() {
            throw new UnsupportedOperationException(
                    "The selector() method may not be called for this type of selection key.");
        }

        @Override
        public boolean isValid() {
            return wrappedSelectionKey.isValid();
        }

        @Override
        public void cancel() {
            wrappedSelectionKey.cancel();
        }

        @Override
        public int interestOps() {
            return wrappedSelectionKey.interestOps();
        }

        @Override
        public SelectionKey interestOps(int ops) {
            wrappedSelectionKey.interestOps(ops);
            return this;
        }

        @Override
        public int readyOps() {
            return wrappedSelectionKey.readyOps();
        }

    }

    private static final Object DUMMY_OBJECT = new Object();

    private final Object channelRegistrationLock = new Object();

    private ThreadLocal<Object> inControllerThread = new ThreadLocal<Object>();
    private ConcurrentLinkedQueue<TimerEvent> newTimerQueue = new ConcurrentLinkedQueue<TimerEvent>();
    private Selector selector;
    private PriorityQueue<TimerEvent> timerQueue = new PriorityQueue<TimerEvent>(
            11, new Comparator<TimerEvent>() {
                @Override
                public int compare(TimerEvent event1, TimerEvent event2) {
                    if (event1.eventTime > event2.eventTime) {
                        return 1;
                    } else if (event1.eventTime < event2.eventTime) {
                        return -1;
                    } else {
                        return 0;
                    }
                }
            });

    /**
     * Creates a communication controller using the platform's default
     * {@link SelectorProvider}.
     * 
     * @throws IOException
     *             if {@link SelectorProvider#openSelector()} throws such an
     *             exception.
     */
    public SimpleCommunicationController() throws IOException {
        this(SelectorProvider.provider());
    }

    /**
     * Creates a communication controller that uses the specified selector
     * provider for opening its {@link Selector}.
     * 
     * @param selectorProvider
     *            selector provider to use for creating this controller's
     *            selector.
     * @throws IOException
     *             if {@link SelectorProvider#openSelector()} throws such an
     *             exception.
     */
    public SimpleCommunicationController(SelectorProvider selectorProvider)
            throws IOException {
        this.inControllerThread = new ThreadLocal<Object>();
        this.selector = selectorProvider.openSelector();
    }

    /**
     * Processes all timers that are due an then waits for an I/O event to
     * happen or a timer to expire. Any exception thrown by
     * {@link ChannelProcessor#processIO()} or
     * {@link TimerProcessor#processTimer()} is not caught by this method, but
     * passed up the stack.
     * 
     * @throws IOException
     *             if {@link Selector#select()} throws such an exception.
     */
    public void processTimersAndIO() throws IOException {
        // We set a marker so that we know when one of this class's methods is
        // called from within the same thread. This allows for some
        // simplifications regarding use of the selector.
        inControllerThread.set(DUMMY_OBJECT);
        try {
            addNewTimerEvents();
            long waitTime = processTimerEvents();
            synchronized (channelRegistrationLock) {
                // We shortly acquire the channel registration lock. This
                // ensures that select(...) will not be called while another
                // thread wants to register a channel.
            }
            selector.select(waitTime);
            // We remember the processors that we already processed so that we
            // do not process processor twice if it has registered multiple
            // channels.
            IdentityHashMap<ChannelProcessor, Void> seenProcessors = new IdentityHashMap<ChannelProcessor, Void>();
            for (Iterator<SelectionKey> selectionKeyIterator = selector
                    .selectedKeys().iterator(); selectionKeyIterator.hasNext();) {
                SelectionKey selectionKey = selectionKeyIterator.next();
                // We can safely cast here, because we know that we attach
                // CommunicationProcessor objects to all selection keys.
                ChannelProcessor processor = (ChannelProcessor) selectionKey
                        .attachment();
                selectionKeyIterator.remove();
                if (!seenProcessors.containsKey(processor)) {
                    seenProcessors.put(processor, null);
                    // If the processIO method throws an exception, it will be
                    // passed up to the method that called this method.
                    processor.processIO();
                }
            }
        } finally {
            // This thread is no longer within this class's control logic, so we
            // reset the marker.
            inControllerThread.remove();
        }
    }

    private void addNewTimerEvents() {
        // We remove all events from the new timer queue and add them to the
        // priority queue so that they are ordered correctly.
        TimerEvent event;
        while ((event = newTimerQueue.poll()) != null) {
            timerQueue.add(event);
        }
    }

    private long processTimerEvents() {
        // We have to remove the event from the timer queue instead of just
        // peeking because another thread might add an event with a smaller
        // event time and then we would remove the wrong event later.
        TimerEvent nextEvent;
        // We use an identity hash-map for remembering which processors we have
        // already seen. This way, we can avoid processing the same processor
        // twice when there are multiple expired timers for it.
        IdentityHashMap<TimerProcessor, Void> seenProcessors = new IdentityHashMap<TimerProcessor, Void>();
        // We use the same "current" time for all timers. This ensures that we
        // do not process a timer for a processor and then ignore future timers
        // that actually expired after the first timer was processed.
        long currentTime = System.currentTimeMillis();
        while ((nextEvent = timerQueue.poll()) != null) {
            if (currentTime >= nextEvent.getEventTime()) {
                // We only process the timer if its processor has not been
                // processed yet.
                TimerProcessor processor = nextEvent.getProcessor();
                // We only process this event if no event for this processor has
                // been processed yet.
                if (!seenProcessors.containsKey(processor)) {
                    seenProcessors.put(processor, null);
                    // If the processTimer method throws an exception, it will
                    // be passed all the way up to the method that called this
                    // objects's processTimersAndIO method.
                    processor.processTimer();
                }
            } else {
                // The time for the next event has not come yet, this means that
                // we can put the event back on the queue and wait.
                timerQueue.add(nextEvent);
                // We update the current time and check again, whether the
                // event's timer has expired. When updating the current time, we
                // also have to clear the set of seen communication processors
                // because we might have processed them before the timers that
                // might be due now even though they were not due back then.
                currentTime = System.currentTimeMillis();
                seenProcessors.clear();
                long nextEventTime = nextEvent.getEventTime();
                if (currentTime < nextEventTime) {
                    // The next event is still not due, so we can wait some time
                    // until the next event is due.
                    return nextEventTime - currentTime;
                }
            }
        }
        // There are no events left in the queue, therefore we can wait
        // indefinitely.
        return 0;
    }

    /**
     * <p>
     * Wakes up the currently waiting {@link #processTimerEvents()} method. If
     * this method is currently not waiting, it will return immediately instead
     * of waiting the next time it is called.
     * </p>
     * 
     * <p>
     * Registering a timer with a start time in the past should be preferred
     * over calling this method. Due to limitations of the Java platform, this
     * method does not establish a happens-before relationship between the code
     * calling this method and the code that is executed in the communication
     * thread after waking up. On many platforms, such a happens-before
     * relationship might exist implicitly, however it is not guaranteed by the
     * API documentation.
     * </p>
     * 
     * <p>
     * Code that wants to trigger execution of a callback within the
     * communication thread should prefer the
     * {@link #registerTimer(TimerProcessor, long)} method. This guarantees a
     * happens-before relationship and will also wake up the communication
     * thread if it is currently sleeping.
     * </p>
     */
    public void wakeUp() {
        selector.wakeup();
    }

    @Override
    public SelectionKey registerChannel(
            final ChannelProcessor channelProcessor,
            final SelectableChannel channel) throws ClosedChannelException {
        if (channelProcessor == null) {
            throw new NullPointerException();
        }
        if (inCommunicationThread()) {
            // We do not have to call wakeup because we are already in the right
            // thread.
            return new WrappingSelectionKey(channel.register(selector, 0,
                    channelProcessor));
        } else {
            synchronized (channelRegistrationLock) {
                // We wake up the selector. If we did not, the register(...)
                // method could block until the current select(...) finished,
                // which could take a long time or even forever.
                selector.wakeup();
                // Now we can safely call register(...). We hold the channel
                // registration lock and the the thread that calls select(...)
                // shortly acquires this lock before starting the select
                // operation.
                return new WrappingSelectionKey(channel.register(selector, 0,
                        channelProcessor));
            }
        }
    }

    @Override
    public void registerTimer(TimerProcessor timerProcessor, long eventTime) {
        if (timerProcessor == null) {
            throw new NullPointerException();
        }
        // We add the timer event to a concurrent queue. The communication
        // thread check this queue and adds the events to the priority queue to
        // put them in the right order.
        newTimerQueue.add(new TimerEvent(eventTime, timerProcessor));
        // We have to wake up the selector because the event we just added might
        // be due before it wakes up by itself.
        selector.wakeup();
    }

    @Override
    public boolean inCommunicationThread() {
        return inControllerThread.get() != null;
    }

}
