/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

/**
 * Collects constants needed for the Channel Access protocol.
 * 
 * @see ChannelAccessCommand
 * @see ChannelAccessVersion
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessConstants {

    /**
     * Minimum (lowest) priority for a Channel Access TCP connection.
     */
    public final static short MIN_PRIORITY = 0;

    /**
     * Maximum (highest) priority for a Channel Access TCP connection.
     */
    public final static short MAX_PRIORITY = 99;

    /**
     * Size of a standard Channel Access header (in bytes). This header is used
     * by most messages unless there is a large payload.
     * 
     * @see #LARGE_HEADER_SIZE
     */
    public final static int STANDARD_HEADER_SIZE = 16;

    /**
     * Size of a large Channel Access header (in bytes). This header is only
     * used for messages with a large payload and if the protocol version
     * supports it.
     * 
     * @see #STANDARD_HEADER_SIZE
     */
    public final static int LARGE_HEADER_SIZE = 24;

    /**
     * Number of bytes a Channel Access message must be aligned to (including
     * the header). As the header size is a multiple of this number, a header
     * only message is always aligned without extra padding.
     */
    public final static int MESSAGE_SIZE_ALIGNMENT = 8;

    /**
     * <p>
     * Size limit for a channel name. We could fit a longer name into a message,
     * however, if the message is transferred using UDP, the packet size is
     * rather limited. Due to limitations of the EPICS IOC, channel names that
     * are longer than about 100 bytes rarely exist anyway. The size of 488
     * bytes (including the terminating null-byte) ensures that at least two
     * search messages fit into a single UDP packet, if the UDP packet has the
     * maximum size of 1024 bytes that is allowed by the protocol.
     * </p>
     * 
     * <p>
     * Please note that there is no guarantee that a message with a channel name
     * of this size can be send via UDP: If the UDP send buffer is smaller, the
     * maximum UDP packet size and thus the limit for the maximum length of a
     * channel name can be smaller.
     * </p>
     */
    public final static int MAX_CHANNEL_NAME_BYTES_LENGTH = 488;

    /**
     * Default TCP and UDP port used by servers. This is the port on which an
     * IOC or name server typically listens for incoming name-resolution
     * requests and on which an IOC typically listens for incoming connections
     * to channels.
     */
    public final static int DEFAULT_SERVER_PORT = 5064;

    /**
     * Default UDP port used by repeaters. This is the port on which a repeater
     * typically listens for incoming beacons from IOCs and registration
     * requests from clients.
     */
    public final static int DEFAULT_REPEATER_PORT = 5065;

    /**
     * Default constructor. This constructor is declared private, so that no
     * instance of this class can be created.
     */
    private ChannelAccessConstants() {
    };

}
