/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;

/**
 * Representation of a <code>CA_PROTO_NOT_FOUND</code> message. This message is
 * sent from a server to a client via UDP in order to indicate that a channel
 * that the client has been looking for could not be found on this server. This
 * message should only be sent if explicitly requested by a client and if the
 * the search request was not sent to a broadcast address.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_NOT_FOUND
 */
public final class ChannelAccessNotFoundMessage extends ChannelAccessMessage {

    private int cid;
    private short dataType;
    private int count;
    private int contextSpecific;

    /**
     * Constructor for creating a <code>CA_PROTO_NOT_FOUND</code> message.
     * 
     * @param searchMessage
     *            the search message in response to which this message is sent
     *            as a reply.
     */
    public ChannelAccessNotFoundMessage(
            ChannelAccessSearchClientMessage searchMessage) {
        super(ChannelAccessCommand.CA_PROTO_NOT_FOUND);

        if (searchMessage == null) {
            throw new NullPointerException(
                    "Search message parameter must not be null.");
        }

        if (searchMessage.hasOriginalMessageHeader()) {
            ChannelAccessMessageHeader searchMessageHeader = searchMessage
                    .getOriginalMessageHeader();
            this.cid = searchMessageHeader.getCID();
            this.dataType = searchMessageHeader.getDataType();
            this.count = searchMessageHeader.getCount();
            this.contextSpecific = searchMessageHeader.getContextSpecific();
        } else {
            this.cid = searchMessage.getChannelCID();
            this.dataType = (short) (searchMessage.isDoReply() ? 10 : 5);
            this.count = searchMessage.getVersion().getMinorVersionAsShort();
            this.contextSpecific = this.cid;
        }
    }

    private ChannelAccessNotFoundMessage(
            ChannelAccessMessageHeader messageHeader) {
        super(ChannelAccessCommand.CA_PROTO_NOT_FOUND, messageHeader);

        this.cid = messageHeader.getCID();
        this.dataType = messageHeader.getDataType();
        this.count = messageHeader.getCount();
        this.contextSpecific = messageHeader.getContextSpecific();
    }

    /**
     * Returns the channel ID that is used by the client to identify the
     * channel. This integer may be zero or negative.
     * 
     * @return client-side ID for the channel.
     */
    public int getChannelCID() {
        return contextSpecific;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, dataType, count, cid, contextSpecific,
                version, maxPayloadSize);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        verifyHeader(0, count, version, maxPayloadSize);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessNotFoundMessage other = (ChannelAccessNotFoundMessage) obj;
        return new EqualsBuilder().append(this.cid, other.cid)
                .append(this.dataType, other.dataType)
                .append(this.count, other.count)
                .append(this.contextSpecific, other.contextSpecific).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode()).append(cid)
                .append(dataType).append(count).append(contextSpecific)
                .toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("cid=");
        sb.append(cid);
        sb.append(", dataType=");
        sb.append(dataType);
        sb.append(", count=");
        sb.append(count);
        sb.append(", contextSpecific=");
        sb.append(contextSpecific);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete message (including the payload).
     */
    protected static ChannelAccessNotFoundMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_NOT_FOUND
                .getCommandNumber());
        byteSource.skip(messageHeader.getPayloadSize());
        return new ChannelAccessNotFoundMessage(messageHeader);
    }

}
