/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.SerializedPayloadTooLargeException;
import com.aquenos.epics.jackie.common.exception.ShortPayloadException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueCodec;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Representation of a <code>CA_PROTO_READ</code> message that is sent from a
 * server to a client. This message is sent in response to a
 * {@link ChannelAccessReadClientMessage} from a client.}.
 * 
 * @see ChannelAccessCommand#CA_PROTO_READ
 * @see ChannelAccessReadMessage
 * @see ChannelAccessReadClientMessage
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessReadServerMessage extends
        ChannelAccessReadMessage {

    private static byte[] PADDING_BYTES = new byte[ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT];

    private short dataType;
    private int count;
    private int cid;
    private int operationId;
    private ChannelAccessGettableValue<?> value;
    private byte[] rawValue;

    /**
     * Creates a <code>CA_PROTO_READ</code> message. This constructor should be
     * used to create a message that is supposed to be sent from a server to a
     * client.
     * 
     * @param dataType
     *            data type of the value in the message's payload (never
     *            <code>null</code>). This should always be the type that the
     *            client requested. If the type requested by the client is not
     *            supported by the server, an error message should be sent
     *            instead of this message.
     * @param count
     *            number of value elements in the message's payload. This should
     *            always be the number of elements that the client requested. If
     *            the number requested by the client is not supported by the
     *            server, an error message should be sent instead of this
     *            message. Must not be negative.
     * @param cid
     *            client-assigned ID identifying the channel.
     * @param operationId
     *            client-assigned ID identifying the read request. This must be
     *            the ID sent with the read request so that the client can
     *            identify the request to which the reply belongs.
     * @param value
     *            current value of the channel (never <code>null</code>).
     * @throws IllegalArgumentException
     *             if the <code>dataType</code> does not match the
     *             <code>value</code>'s data type or <code>count</code> is
     *             negative or greater than the <code>value</code>'s number of
     *             elements.
     */
    public ChannelAccessReadServerMessage(ChannelAccessValueType dataType,
            int count, int cid, int operationId,
            ChannelAccessGettableValue<?> value) {
        super();

        if (!dataType.equals(value.getType())) {
            throw new IllegalArgumentException("The specified data type ("
                    + dataType.toString()
                    + ") does not match the value's data type ("
                    + value.getType() + ").");
        }
        if (count < 0) {
            throw new IllegalArgumentException(
                    "The count must not be negative.");
        }
        if (count > value.getValueSize()) {
            throw new IllegalArgumentException("The count (" + count
                    + ") exceeds the number of elements in the value ("
                    + value.getValueSize() + ").");
        }
        this.dataType = dataType.toTypeCode();
        this.count = count;
        this.cid = cid;
        this.operationId = operationId;
        this.value = value.asReadOnlyValue();
    }

    private ChannelAccessReadServerMessage(
            ChannelAccessMessageHeader messageHeader, short dataType,
            int count, int cid, int operationId,
            ChannelAccessGettableValue<?> value, byte[] rawValue) {
        super(messageHeader);

        this.dataType = dataType;
        this.count = count;
        this.cid = cid;
        this.operationId = operationId;
        if (value != null) {
            this.value = value.asReadOnlyValue();
        }
        this.rawValue = rawValue;
    }

    /**
     * Returns the data type of the value in this message's payload. This should
     * always match the data type that the client requested. Returns
     * <code>null</code> if the specified data-type is unknown to this
     * implementation.
     * 
     * @return data-type of the value or <code>null</code> if the numeric
     *         identifier cannot be mapped to a data type known by this
     *         implementation.
     */
    public ChannelAccessValueType getDataType() {
        try {
            ChannelAccessValueType type = ChannelAccessValueType
                    .forTypeNumber(this.dataType);
            // We treat a data type that is not valid for this message the same
            // way we treat an unknown data type.
            if (type.isGettable()) {
                return type;
            } else {
                return null;
            }
        } catch (IllegalArgumentException e) {
            return null;
        }
    }

    /**
     * Returns the number of elements the value in this message's payload has.
     * This should always match the element count that the client requested.
     * 
     * @return number of value elements.
     */
    public int getCount() {
        return count;
    }

    /**
     * Returns the channel CID used by the client. This is the identifier that
     * the client uses to identify the channel. This number might be negative or
     * zero.
     * 
     * @return channel CID assigned to the respective channel by the client.
     */
    public int getChannelCID() {
        return cid;
    }

    /**
     * Returns the client-assigned operation ID. The operation ID is used by the
     * client in order to map replies sent in response to a read request to the
     * respective request. This must always be the ID that was sent by the
     * client. This number might be negative or zero.
     * 
     * @return client-assigned operation identifier.
     */
    public int geOperationId() {
        return operationId;
    }

    /**
     * Returns the current value of the channel. This method returns
     * <code>null</code> if the data-type sent by the server is not supported by
     * this implementation ({@link #getDataType()} returns <code>null</code>).
     * 
     * @return current value of the channel or <code>null</code> if the data
     *         type is not supported by this implementation. The value's data
     *         type is the same that is returned by {@link #getDataType()} and
     *         the value's element count is equal to or greater than the number
     *         returned by {@link #getCount()}.
     */
    public ChannelAccessGettableValue<?> getValue() {
        if (value == null) {
            return null;
        } else {
            return value;
        }
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        // maxPayloadSize should be aligned.
        assert (maxPayloadSize % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT == 0);
        ChannelAccessValueType dataTypeEnum = ChannelAccessValueType
                .forTypeNumber(dataType);
        if (value != null) {
            // This is the usual case. The data-type is valid and we have a
            // value.
            int maxCount = ChannelAccessValueCodec.calculateMaxCount(
                    dataTypeEnum, maxPayloadSize);
            if (count > maxCount || count < 0) {
                throw new SerializedPayloadTooLargeException(
                        "Cannot send message with value type "
                                + dataTypeEnum.toString() + " and count "
                                + (count & 0xffffffffL)
                                + " because payload size limit is set to "
                                + maxPayloadSize + ".");
            }
            int payloadSize = ChannelAccessValueCodec.calculatePayloadSize(
                    ChannelAccessValueType.forTypeNumber(dataType), count);
            int paddingSize = calculatePaddingSize(payloadSize);
            serializeHeader(byteSink, payloadSize + paddingSize, dataType,
                    count, cid, operationId, version, maxPayloadSize);
            ChannelAccessValueCodec.encodeGettableValue(byteSink, value, count);
            byteSink.putByteArray(PADDING_BYTES, 0, paddingSize);
        } else {
            // We do not have a value. This means that we could not handle the
            // type and thus we could not deserialize the value. However, we can
            // simply use the raw value.
            serializeHeader(byteSink, rawValue.length, dataType, count, cid,
                    operationId, version, maxPayloadSize);
            byteSink.putByteArray(rawValue);
        }
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        // maxPayloadSize should be aligned.
        assert (maxPayloadSize % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT == 0);
        ChannelAccessValueType dataTypeEnum = ChannelAccessValueType
                .forTypeNumber(dataType);
        if (value != null) {
            // This is the usual case. The data-type is valid and we have a
            // value.
            int maxCount = ChannelAccessValueCodec.calculateMaxCount(
                    dataTypeEnum, maxPayloadSize);
            if (count > maxCount || count < 0) {
                throw new SerializedPayloadTooLargeException(
                        "Cannot send message with value type "
                                + dataTypeEnum.toString() + " and count "
                                + (count & 0xffffffffL)
                                + " because payload size limit is set to "
                                + maxPayloadSize + ".");
            }
            int payloadSize = ChannelAccessValueCodec.calculatePayloadSize(
                    ChannelAccessValueType.forTypeNumber(dataType), count);
            int paddingSize = calculatePaddingSize(payloadSize);
            verifyHeader(payloadSize + paddingSize, count, version,
                    maxPayloadSize);
        } else {
            // We do not have a value. This means that we could not handle the
            // type and thus we could not deserialize the value. However, we can
            // simply use the raw value.
            verifyHeader(rawValue.length, count, version, maxPayloadSize);
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessReadServerMessage other = (ChannelAccessReadServerMessage) obj;
        return new EqualsBuilder().append(this.dataType, other.dataType)
                .append(this.count, other.count).append(this.cid, other.cid)
                .append(this.operationId, other.operationId)
                .append(this.value, other.value)
                .append(this.rawValue, other.rawValue).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(dataType).append(count).append(cid).append(operationId)
                .append(value).append(rawValue).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("dataType=");
        try {
            sb.append(ChannelAccessValueType.forTypeNumber(dataType));
        } catch (IllegalArgumentException e) {
            sb.append("UNKNOWN(");
            sb.append(dataType);
            sb.append(')');
        }
        sb.append(", count=");
        sb.append(count);
        sb.append(", cid=");
        sb.append(cid);
        sb.append(", operationId=");
        sb.append(operationId);
        if (value != null) {
            sb.append(", value=");
            sb.append(value);
        }
        if (rawValue != null) {
            sb.append(", rawValue=");
            addByteArrayToStringBuilder(sb, rawValue);
        }
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param charset
     *            encoding to use for deserializing the host name. Usually, host
     *            names should consist of ASCII characters only, however this is
     *            not strictly enforced by the protocol, so the encoding might
     *            matter.
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     * @throws ShortPayloadException
     *             if the message's payload size is less than the size needed
     *             for the data type and count.
     */
    protected static ChannelAccessReadServerMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            Charset charset) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_READ
                .getCommandNumber());

        short dataTypeNumber = messageHeader.getDataType();
        ChannelAccessValueType dataType;
        try {
            dataType = ChannelAccessValueType.forTypeNumber(dataTypeNumber);
            // We treat a data type that is not valid for this message the same
            // way we treat an unknown data type.
            if (!dataType.isGettable()) {
                dataType = null;
            }
        } catch (IllegalArgumentException e) {
            dataType = null;
        }
        int count = messageHeader.getCount();
        int cid = messageHeader.getCID();
        int subscriptionId = messageHeader.getContextSpecific();
        int actualPayloadSize = messageHeader.getPayloadSize();
        ChannelAccessGettableValue<?> value;
        byte[] rawValue;
        if (dataType == null) {
            value = null;
            rawValue = byteSource.getByteArray(actualPayloadSize);
        } else {
            int expectedPayloadSize;
            try {
                // calculatePayloadSize throws an IllegalArgumentException
                // if count is negative or too large to represent the
                // resulting number as an int.
                expectedPayloadSize = ChannelAccessValueCodec
                        .calculatePayloadSize(dataType, count);
            } catch (IllegalArgumentException e) {
                throw new ShortPayloadException("Payload size of "
                        + actualPayloadSize + " is too small to hold a "
                        + dataType.toString() + " with "
                        + (count & 0xffffffffL) + " elements.");
            }
            if (actualPayloadSize < expectedPayloadSize) {
                throw new ShortPayloadException(
                        "Received a CA_PROTO_READ message with a payload size of "
                                + actualPayloadSize
                                + ". However, the message should have a payload of at least "
                                + expectedPayloadSize + " bytes.");
            }
            value = ChannelAccessValueCodec.decodeGettableValue(byteSource,
                    dataType, expectedPayloadSize, count, charset);
            byteSource.skip(actualPayloadSize - expectedPayloadSize);
            rawValue = null;
        }
        return new ChannelAccessReadServerMessage(messageHeader,
                dataTypeNumber, count, cid, subscriptionId, value, rawValue);
    }

}
