/*
 * Copyright 2016-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.DatagramChannel;

/**
 * Utility methods for dealing with {@link DatagramChannel}s.
 * 
 * @author Sebastian Marsching
 */
public final class DatagramChannelUtil {

    private DatagramChannelUtil() {
    }

    private static final Method DATAGRAM_CHANNEL_OPEN_METHOD;
    private static final Method DATAGRAM_CHANNEL_SET_OPTION_METHOD;
    private static final boolean MULTICAST_AVAILABLE;
    private static final boolean PROTOCOL_FAMILY_AVAILABLE;
    private static final Object PROTOCOL_FAMILY_INET;
    private static final Object SOCKET_OPTION_IP_MULTICAST_LOOP;
    private static final Object SOCKET_OPTION_IP_MULTICAST_TTL;

    static {
        Object protocolFamilyInet;
        Method datagramChannelOpenMethod;
        Method datagramChannelSetOptionMethod;
        Object socketOptionIpMulticastLoop;
        Object socketOptionIpMulticastTtl;
        // We try to get the API objects needed for setting the protocol family
        // and configuring multicast options. We do both in a single try block
        // because both features are supported since Java 7, so they should
        // either both succeed or both fail.
        try {
            Class<?> protocolFamilyClass = Class
                    .forName("java.net.ProtocolFamily");
            Class<?> standardProtocolFamilyClass = Class
                    .forName("java.net.StandardProtocolFamily");
            protocolFamilyInet = standardProtocolFamilyClass.getField("INET")
                    .get(null);
            datagramChannelOpenMethod = DatagramChannel.class.getMethod("open",
                    protocolFamilyClass);
            // We can only use the DatagramChannel.open(...) method, if
            // StandardProtocolFamily implements ProtocolFamily (and thus we can
            // pass StandardProtocolFamily.INET to the open(...) method, if the
            // method is public and static, and if the method returns a
            // DatagramChannel.
            if (!protocolFamilyClass
                    .isAssignableFrom(standardProtocolFamilyClass)
                    || (datagramChannelOpenMethod.getModifiers()
                            & (Modifier.PUBLIC
                                    | Modifier.STATIC)) != (Modifier.PUBLIC
                                            | Modifier.STATIC)
                    || !DatagramChannel.class.isAssignableFrom(
                            datagramChannelOpenMethod.getReturnType())) {
                protocolFamilyInet = null;
                datagramChannelOpenMethod = null;
            }
            Class<?> socketOptionClass = Class.forName("java.net.SocketOption");
            datagramChannelSetOptionMethod = DatagramChannel.class
                    .getMethod("setOption", socketOptionClass, Object.class);
            Class<?> standardSocketOptionsClass = Class
                    .forName("java.net.StandardSocketOptions");
            socketOptionIpMulticastLoop = standardSocketOptionsClass
                    .getField("IP_MULTICAST_LOOP").get(null);
            socketOptionIpMulticastTtl = standardSocketOptionsClass
                    .getField("IP_MULTICAST_TTL").get(null);
            // We can only use the DatagramChannel.setSocketOption(...) method,
            // if the socket options are compatible with the type of the
            // method's parameter and if the socket options expect a value of
            // the right type.
            Method socketOptionTypeMethod = socketOptionClass.getMethod("type");
            if (!socketOptionClass.isInstance(socketOptionIpMulticastLoop)
                    || !socketOptionClass.isInstance(socketOptionIpMulticastTtl)
                    || (datagramChannelSetOptionMethod.getModifiers()
                            & Modifier.PUBLIC) != Modifier.PUBLIC
                    || socketOptionTypeMethod.invoke(
                            socketOptionIpMulticastLoop) != Boolean.class
                    || socketOptionTypeMethod.invoke(
                            socketOptionIpMulticastTtl) != Integer.class) {
                datagramChannelOpenMethod = null;
                socketOptionIpMulticastLoop = null;
                socketOptionIpMulticastTtl = null;
            }
        } catch (Throwable t) {
            protocolFamilyInet = null;
            datagramChannelOpenMethod = null;
            datagramChannelSetOptionMethod = null;
            socketOptionIpMulticastLoop = null;
            socketOptionIpMulticastTtl = null;
        }
        PROTOCOL_FAMILY_INET = protocolFamilyInet;
        DATAGRAM_CHANNEL_OPEN_METHOD = datagramChannelOpenMethod;
        PROTOCOL_FAMILY_AVAILABLE = PROTOCOL_FAMILY_INET != null
                && DATAGRAM_CHANNEL_OPEN_METHOD != null;
        DATAGRAM_CHANNEL_SET_OPTION_METHOD = datagramChannelSetOptionMethod;
        SOCKET_OPTION_IP_MULTICAST_LOOP = socketOptionIpMulticastLoop;
        SOCKET_OPTION_IP_MULTICAST_TTL = socketOptionIpMulticastTtl;
        MULTICAST_AVAILABLE = DATAGRAM_CHANNEL_SET_OPTION_METHOD != null
                && SOCKET_OPTION_IP_MULTICAST_LOOP != null
                && SOCKET_OPTION_IP_MULTICAST_TTL != null;
    }

    /**
     * <p>
     * Opens a new {@link DatagramChannel}. If possible, this method will call
     * <code>DatagramChannel.open(ProtocolFamily)</code>, passing
     * <code>StandardProtocolFamily.INET</code>. This will limit the socket to
     * IPv4.
     * </p>
     * 
     * <p>
     * Unfortunately, the <code>DatagramChannel.open(ProtocolFamily)</code>
     * method is only available in JDK 1.7 and newer. In order to stay
     * compatible with JDK 1.6 (both at compile-time and run-time), this method
     * actually uses reflection for calling the
     * <code>DatagramChannel.open(ProtocolFamily)</code> method. If the method
     * cannot be called (for example because the software is running on JRE 1.6
     * or reflection fails due to security restrictions), it falls back to
     * calling {@link DatagramChannel#open()}. In this case, however, the
     * returned channel will not be restricted to IPv4.
     * </p>
     * 
     * <p>
     * Any exception thrown by <code>DatagramChannel.open()</code> or
     * <code>DatagramChannel.open(ProtocolFamily)</code> is also thrown by this
     * method. Any exception caused by reflection is caught by this method and
     * {@link DatagramChannel#open()} is called directly instead.
     * </p>
     * 
     * @return datagram channel returned by <code>DatagramChannel.open()</code>
     *         or <code>DatagramChannel.open(StandardProtocolFamily.INET)</code>
     *         (the latter is preferred).
     * @throws IOException
     *             if the called <code>open</code> method throws such an
     *             exception.
     */
    public static DatagramChannel openChannelInet4() throws IOException {
        if (PROTOCOL_FAMILY_AVAILABLE) {
            try {
                return (DatagramChannel) DATAGRAM_CHANNEL_OPEN_METHOD
                        .invoke(null, PROTOCOL_FAMILY_INET);
            } catch (InvocationTargetException e) {
                // If the exception was caused by DatagramChannel.open(...)
                // throwing an exception, we want to rethrow it. If the cause is
                // not of one of the types thrown by DatagramChannel.open(...),
                // we fall-through to using the regular DatagramChannel.open()
                // method.
                Throwable cause = e.getCause();
                if (cause instanceof IOException) {
                    throw (IOException) cause;
                } else if (cause instanceof RuntimeException) {
                    throw (RuntimeException) cause;
                } else if (cause instanceof Error) {
                    throw (Error) cause;
                }
            } catch (Exception e) {
                // If the invocation fails for any reason that is caused by
                // reflection (and not by the actual method), we fall-through to
                // using the regular DatagramChannel.open() method.
            }
        }
        return DatagramChannel.open();
    }

    /**
     * <p>
     * Sets the IP multicast loop option on a datagram channel.
     * </p>
     * 
     * <p>
     * When running on Java 7, this method calls
     * <code>DatagramChannel.setOption(StandardSocketOptions.IP_MULTICAST_LOOP, loopBack)</code>.
     * When running on older versions of Java (where the <code>setOption</code>
     * method and the <code>StandardSocketOptions</code> class are missing, this
     * method does nothing.
     * </p>
     * 
     * @param channel
     *            channel on which the <code>setOption</code> method should be
     *            called.
     * @param loopBack
     *            new value for the <code>IP_MULTICAST_LOOP</code> flag.
     * @throws ClosedChannelException
     *             if <code>setOption</code> throws such an exception.
     * @throws IllegalArgumentException
     *             if <code>setOption</code> throws such an exception.
     * @throws IOException
     *             if <code>setOption</code> throws such an exception.
     * @throws UnsupportedOperationException
     *             if <code>setOption</code> throws such an exception.
     */
    public static void setIpMulticastLoopOption(DatagramChannel channel,
            boolean loopBack) throws IOException {
        // Java 6 does not support setting multicast options, so we do not do
        // anything on that platform.
        if (!MULTICAST_AVAILABLE) {
            return;
        }
        try {
            DATAGRAM_CHANNEL_SET_OPTION_METHOD.invoke(null,
                    SOCKET_OPTION_IP_MULTICAST_LOOP, loopBack);
        } catch (InvocationTargetException e) {
            // If the exception was caused by
            // DatagramChannel.setSocketOption(...) throwing an exception,
            // we want to rethrow it. If the cause is not of one of the
            // types thrown by DatagramChannel.open(...),
            // we fall-through to using the regular DatagramChannel.open()
            // method.
            Throwable cause = e.getCause();
            if (cause instanceof IOException) {
                throw (IOException) cause;
            } else if (cause instanceof RuntimeException) {
                throw (RuntimeException) cause;
            } else if (cause instanceof Error) {
                throw (Error) cause;
            }
        } catch (Exception e) {
            // If the invocation fails for any reason that is caused by
            // reflection (and not by the actual method), we do not want
            // this exception to bubble up. Most likely, it was simply
            // caused because the JRE we are running on is too old.
        }
    }

    /**
     * <p>
     * Sets the IP multicast ttl option on a datagram channel.
     * </p>
     * 
     * <p>
     * When running on Java 7, this method calls
     * <code>DatagramChannel.setOption(StandardSocketOptions.IP_MULTICAST_TTL, ttl)</code>.
     * When running on older versions of Java (where the <code>setOption</code>
     * method and the <code>StandardSocketOptions</code> class are missing, this
     * method does nothing.
     * </p>
     * 
     * @param channel
     *            channel on which the <code>setOption</code> method should be
     *            called.
     * @param ttl
     *            new value for the <code>IP_MULTICAST_TTL</code> option.
     * @throws ClosedChannelException
     *             if <code>setOption</code> throws such an exception.
     * @throws IllegalArgumentException
     *             if <code>setOption</code> throws such an exception.
     * @throws IOException
     *             if <code>setOption</code> throws such an exception.
     * @throws UnsupportedOperationException
     *             if <code>setOption</code> throws such an exception.
     */
    public static void setIpMulticastTtlOption(DatagramChannel channel, int ttl)
            throws IOException {
        // Java 6 does not support setting multicast options, so we do not do
        // anything on that platform.
        if (!MULTICAST_AVAILABLE) {
            return;
        }
        try {
            DATAGRAM_CHANNEL_SET_OPTION_METHOD.invoke(null,
                    SOCKET_OPTION_IP_MULTICAST_TTL, ttl);
        } catch (InvocationTargetException e) {
            // If the exception was caused by
            // DatagramChannel.setSocketOption(...) throwing an exception,
            // we want to rethrow it. If the cause is not of one of the
            // types thrown by DatagramChannel.open(...),
            // we fall-through to using the regular DatagramChannel.open()
            // method.
            Throwable cause = e.getCause();
            if (cause instanceof IOException) {
                throw (IOException) cause;
            } else if (cause instanceof RuntimeException) {
                throw (RuntimeException) cause;
            } else if (cause instanceof Error) {
                throw (Error) cause;
            }
        } catch (Exception e) {
            // If the invocation fails for any reason that is caused by
            // reflection (and not by the actual method), we do not want
            // this exception to bubble up. Most likely, it was simply
            // caused because the JRE we are running on is too old.
        }
    }

}
