/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;

import org.junit.Test;

/**
 * Tests for {@link SimpleConcurrentQueue}.
 * 
 * @author Sebastian Marsching
 */
public class SimpleConcurrentQueueTest {

    /**
     * Test the queue from a single thread.
     */
    @Test
    public void singleThread() {
        SimpleConcurrentQueue<Integer> queue = new SimpleConcurrentQueue<Integer>();
        assertNull(queue.peek());
        assertFalse(queue.removeIfMatch(5));
        queue.add(1);
        assertEquals(1, (int) queue.peek());
        assertFalse(queue.removeIfMatch(5));
        assertTrue(queue.removeIfMatch(1));
        queue.add(2);
        queue.add(3);
        queue.add(4);
        assertTrue(queue.removeIfMatch(2));
        assertTrue(queue.removeIfMatch(3));
        assertTrue(queue.removeIfMatch(4));
        assertNull(queue.peek());
    }

    /**
     * Tests that trying to add a <code>null</code> element results in a
     * {@link NullPointerException}.
     */
    @Test(expected = NullPointerException.class)
    public void addNull() {
        SimpleConcurrentQueue<Integer> queue = new SimpleConcurrentQueue<Integer>();
        queue.add(null);
    }

    /**
     * Tests that trying to remove a <code>null</code> element results in a
     * {@link NullPointerException}.
     */
    @Test(expected = NullPointerException.class)
    public void removeNull() {
        SimpleConcurrentQueue<Integer> queue = new SimpleConcurrentQueue<Integer>();
        queue.add(2);
        queue.removeIfMatch(null);
    }

    /**
     * Helper object that is put in the queue for the {@link #parallelThreads()}
     * test.
     */
    private class SequenceObject {
        public boolean endOfSequence;
        public int senderId;
        public int sequenceNumber;
    }

    /**
     * Sender threads for the {@link #parallelThreads()} test.
     */
    private class NumberSender implements Callable<Void> {

        private SimpleConcurrentQueue<SequenceObject> queue;
        private int id;
        private int countFrom;
        private int countTo;

        public NumberSender(SimpleConcurrentQueue<SequenceObject> queue,
                int id, int countFrom, int countTo) {
            this.queue = queue;
            this.id = id;
            this.countFrom = countFrom;
            this.countTo = countTo;
        }

        @Override
        public Void call() throws Exception {
            for (int i = countFrom; i <= countTo; i++) {
                SequenceObject seq = new SequenceObject();
                seq.senderId = id;
                seq.sequenceNumber = i;
                queue.add(seq);
            }
            return null;
        }

    }

    /**
     * Receiver thread for the {@link #parallelThreads()}.
     */
    private class NumberReceiver implements
            Callable<LinkedList<SequenceObject>> {

        private SimpleConcurrentQueue<SequenceObject> queue;
        private int numberOfSenders;

        public NumberReceiver(SimpleConcurrentQueue<SequenceObject> queue,
                int numberOfSenders) {
            this.queue = queue;
            this.numberOfSenders = numberOfSenders;
        }

        @Override
        public LinkedList<SequenceObject> call() throws Exception {
            LinkedList<SequenceObject> receivedObjects = new LinkedList<SequenceObject>();
            int[] senderToLastSeqNumber = new int[numberOfSenders];
            for (;;) {
                SequenceObject seq = queue.peek();
                if (seq != null) {
                    if (queue.removeIfMatch(seq)) {
                        if (seq.endOfSequence) {
                            return receivedObjects;
                        } else {
                            receivedObjects.add(seq);
                            // Check that we do not receive out-of-order
                            // objects.
                            assertTrue(seq.sequenceNumber > senderToLastSeqNumber[seq.senderId]);
                            senderToLastSeqNumber[seq.senderId] = seq.sequenceNumber;
                        }
                    }
                }
            }
        }

    }

    /**
     * Run test with several concurrent threads. This test tries to expose
     * concurrency issues by filling and draining the queue from multiple
     * threads.
     */
    @Test
    public void parallelThreads() throws InterruptedException,
            ExecutionException {
        SimpleConcurrentQueue<SequenceObject> queue = new SimpleConcurrentQueue<SequenceObject>();
        // The combination of 5 senders, 5 receivers and 800000 entries per
        // sender makes a run time of a few seconds on my development computer
        // and thus gives a reasonable chance of detecting concurrency issues
        // while not blocking the test process for too long.
        int numberOfSenders = 5;
        int numberOfReceivers = 5;
        int numberOfSequenceElements = 800000;
        // Create and start sender threads.
        ArrayList<FutureTask<Void>> senders = new ArrayList<FutureTask<Void>>(
                numberOfSenders);
        for (int i = 0; i < numberOfSenders; i++) {
            FutureTask<Void> sender = new FutureTask<Void>(new NumberSender(
                    queue, i, 1, numberOfSequenceElements));
            senders.add(sender);
            new Thread(sender).start();
        }
        // Create and start receiver threads.
        ArrayList<FutureTask<LinkedList<SequenceObject>>> receivers = new ArrayList<FutureTask<LinkedList<SequenceObject>>>(
                numberOfReceivers);
        for (int i = 0; i < numberOfReceivers; i++) {
            FutureTask<LinkedList<SequenceObject>> receiver = new FutureTask<LinkedList<SequenceObject>>(
                    new NumberReceiver(queue, numberOfSenders));
            receivers.add(receiver);
            new Thread(receiver).start();
        }
        // Wait for sender threads to finish.
        for (FutureTask<Void> sender : senders) {
            sender.get();
        }
        // Signal receiver threads to finish.
        SequenceObject endSeq = new SequenceObject();
        endSeq.endOfSequence = true;
        for (int i = 0; i < numberOfReceivers; i++) {
            queue.add(endSeq);
        }
        // Count number of sequence objects received.
        HashSet<SequenceObject> allSequenceObjects = new HashSet<SequenceObject>(
                numberOfSenders * numberOfSequenceElements);
        int totalCount = 0;
        for (FutureTask<LinkedList<SequenceObject>> receiver : receivers) {
            LinkedList<SequenceObject> receivedSequence = receiver.get();
            allSequenceObjects.addAll(receivedSequence);
            totalCount += receivedSequence.size();
        }
        // Check that we received the expected number of sequence objects in
        // total.
        assertEquals(numberOfSenders * numberOfSequenceElements, totalCount);
        // Check that all received objects are distinct.
        assertEquals(numberOfSenders * numberOfSequenceElements,
                allSequenceObjects.size());
    }

}
